# wget https://raw.githubusercontent.com/Sleepwalking/pytorch-softdtw/master/soft_dtw.py
import numpy as np
import torch
from numba import jit
from torch.autograd import Function

@jit(nopython = True)
def compute_softdtw(D, gamma):
  B = D.shape[0]
  N = D.shape[1]
  M = D.shape[2]
  R = np.ones((B, N + 2, M + 2)) * np.inf
  R[:, 0, 0] = 0
  for k in range(B):
    for j in range(1, M + 1):
      for i in range(1, N + 1):
        r0 = -R[k, i - 1, j - 1] / gamma
        r1 = -R[k, i - 1, j] / gamma
        r2 = -R[k, i, j - 1] / gamma
        rmax = max(max(r0, r1), r2)
        rsum = np.exp(r0 - rmax) + np.exp(r1 - rmax) + np.exp(r2 - rmax)
        softmin = - gamma * (np.log(rsum) + rmax)
        R[k, i, j] = D[k, i - 1, j - 1] + softmin
  return R

@jit(nopython = True)
def compute_softdtw_backward(D_, R, gamma):
  B = D_.shape[0]
  N = D_.shape[1]
  M = D_.shape[2]
  D = np.zeros((B, N + 2, M + 2))
  E = np.zeros((B, N + 2, M + 2))
  D[:, 1:N + 1, 1:M + 1] = D_
  E[:, -1, -1] = 1
  R[:, : , -1] = -np.inf
  R[:, -1, :] = -np.inf
  R[:, -1, -1] = R[:, -2, -2]
  for k in range(B):
    for j in range(M, 0, -1):
      for i in range(N, 0, -1):
        a0 = (R[k, i + 1, j] - R[k, i, j] - D[k, i + 1, j]) / gamma
        b0 = (R[k, i, j + 1] - R[k, i, j] - D[k, i, j + 1]) / gamma
        c0 = (R[k, i + 1, j + 1] - R[k, i, j] - D[k, i + 1, j + 1]) / gamma
        a = np.exp(a0)
        b = np.exp(b0)
        c = np.exp(c0)
        E[k, i, j] = E[k, i + 1, j] * a + E[k, i, j + 1] * b + E[k, i + 1, j + 1] * c
  return E[:, 1:N + 1, 1:M + 1]

class _SoftDTW(Function):
  @staticmethod
  def forward(ctx, D, gamma):
    dev = D.device
    dtype = D.dtype
    gamma = torch.Tensor([gamma]).to(dev).type(dtype) # dtype fixed
    D_ = D.detach().cpu().numpy()
    g_ = gamma.item()
    R = torch.Tensor(compute_softdtw(D_, g_)).to(dev).type(dtype)
    ctx.save_for_backward(D, R, gamma)
    return R[:, -2, -2]

  @staticmethod
  def backward(ctx, grad_output):
    dev = grad_output.device
    dtype = grad_output.dtype
    D, R, gamma = ctx.saved_tensors
    D_ = D.detach().cpu().numpy()
    R_ = R.detach().cpu().numpy()
    g_ = gamma.item()
    E = torch.Tensor(compute_softdtw_backward(D_, R_, g_)).to(dev).type(dtype)
    return grad_output.view(-1, 1, 1).expand_as(E) * E, None

class SoftDTW(torch.nn.Module):
  def __init__(self, gamma=1.0, normalize=False):
    super(SoftDTW, self).__init__()
    self.normalize = normalize
    self.gamma=gamma
    self.func_dtw = _SoftDTW.apply

  def calc_distance_matrix(self, x, y):
    n = x.size(1)
    m = y.size(1)
    d = x.size(2)
    x = x.unsqueeze(2).expand(-1, n, m, d)
    y = y.unsqueeze(1).expand(-1, n, m, d)
    dist = torch.pow(x - y, 2).sum(3)
    return dist

  def forward(self, x, y):
    assert x.ndim == y.ndim, f"{x.shape} != {y.shape}"
    squeeze = False
    if x.ndim < 3:
      x = x.unsqueeze(0)
      y = y.unsqueeze(0)
      squeeze = True
    if self.normalize:
      D_xy = self.calc_distance_matrix(x, y)
      out_xy = self.func_dtw(D_xy, self.gamma)
      D_xx = self.calc_distance_matrix(x, x)
      out_xx = self.func_dtw(D_xx, self.gamma)
      D_yy = self.calc_distance_matrix(y, y)
      out_yy = self.func_dtw(D_yy, self.gamma)
      result = out_xy - 1/2 * (out_xx + out_yy) # distance
    else:
      D_xy = self.calc_distance_matrix(x, y)
      out_xy = self.func_dtw(D_xy, self.gamma)
      result = out_xy # discrepancy
    return result.squeeze(0) if squeeze else result



from unittest import TestCase

import torch as t

from pandas_ml_utils_torch.distribution import Logistic


class TestDistribution(TestCase):

    def test_logistic(self):
        dist = Logistic(t.tensor([0.0] * 3), t.tensor([1.0] * 3))
        print(dist.sample())


import os
import unittest
from unittest import TestCase

from pandas_ml_utils import pd, Model
from pandas_ml_common_test.notebook_runner import run_notebook

PWD = os.path.dirname(os.path.abspath(__file__))


class TestSaveAndLoad(TestCase):

    def test_save_load_nb_model(self):
        notebooks_path = os.path.join(PWD, '..', 'examples')
        notebook_file = os.path.join(notebooks_path, 'regression_with_regularization.ipynb')
        out, err = run_notebook(notebook_file, notebooks_path)
        self.assertEqual(err, [])

        df = pd.read_csv(os.path.join(notebooks_path, 'SPY.csv'))
        model = Model.load('/tmp/regression_with_regularization.model')
        prediction = df.model.predict(model, tail=1)["prediction", "Close"].item()

        print(prediction)
        self.assertEqual(float(out["cells"][-1]["outputs"][-1]["data"]["text/plain"]), prediction)


    @unittest.skip("debuging")
    def test_make_model(self):
        notebooks_path = os.path.join(PWD, '..', 'examples')
        df = pd.read_csv(os.path.join(notebooks_path, 'SPY.csv'))

        with df.model("/tmp/pijsfnwuacpa.model") as m:
            from torch import nn
            from torch.optim import SGD
            from pandas_ml_common.utils.column_lagging_utils import lag_columns

            from pandas_ml_utils import FeaturesAndLabels, RegressionSummary, FittingParameter
            from pandas_ml_utils_torch import PytorchModel
            from pandas_ml_utils_torch.merging_cross_folds import take_the_best

            def net_provider():
                from pandas_ml_utils_torch import PytorchNN

                class Net(PytorchNN):

                    def __init__(self):
                        super().__init__()
                        self.net = nn.Sequential(
                            nn.Linear(10, 4),
                            nn.Tanh(),
                            nn.Linear(4, 4),
                            nn.Tanh(),
                            nn.Linear(4, 1),
                            nn.Tanh(),
                        )

                    def L1(self):
                        # path to the parameters which should be regularized
                        # the path is constructed from self.named_parameters() and allows the use of wildcards
                        return {'net/0/**/weight': 0.02}

                    def L2(self):
                        return {
                            'net/0/**/weight': 0.02,
                            'net/2/**/weight': 0.05
                        }

                    def forward_training(self, x):
                        return self.net(x)

                return Net()

            fit = m.fit(
                PytorchModel(
                    net_provider,
                    FeaturesAndLabels(
                        [lambda df: lag_columns(df["Close"].pct_change(), range(10))],
                        [lambda df: df["Close"].pct_change().shift(-1)]),
                    nn.MSELoss,
                    lambda params: SGD(params, lr=0.01, momentum=0.0),
                    merge_cross_folds=take_the_best,
                    summary_provider=RegressionSummary
                ),
                FittingParameter(epochs=2),
                verbose=1
            )

    @unittest.skip("debuging")
    def test_load_only(self):
        from pandas_ml_common.utils.column_lagging_utils import lag_columns
        notebooks_path = os.path.join(PWD, '..', 'examples')
        df = pd.read_csv(os.path.join(notebooks_path, 'SPY.csv'))
        #model = Model.load("/tmp/pijsfnwuacpa.model")
        model = Model.load('/tmp/regression_with_regularization.model')
        prediction = df.model.predict(model, tail=1)

        print(prediction)
        self.assertEqual(len(prediction), 1)



from typing import Union, Tuple, Callable

import torch as t
import torch.nn as nn
from torch.nn import init


class LambdaSplitter(nn.Module):

    def __init__(self,  *funcs: Callable[[t.Tensor], t.Tensor]) -> None:
        super().__init__()
        self.funcs = funcs

    def forward(self, input, *args):
        return tuple([c(input) for c in self.funcs])


class Flatten(nn.Module):

    def __init__(self, *args):
        super().__init__()

    def forward(self, input, *args):
        return input.view(input.size(0), -1)


class Squeeze(nn.Module):

    def __init__(self, *args):
        super().__init__()

    def forward(self, input, *args):
        squoze = input.squeeze()

        # Note that we need to keep the batch dimension in case it is 1
        return squoze if input.shape[0] > 1 else squoze.view(1, *squoze.shape)


class Reshape(nn.Module):

    def __init__(self, *args):
        super().__init__()
        self.shape = args

    def forward(self, x):
        # keep batch size
        return x.view(x.shape[0], *self.shape)


class KerasLikeLSTM(nn.Module):

    def __init__(self, input_shape, output, return_sequence=False, sequential=False):
        super().__init__()
        self.features = input_shape[1] if isinstance(input_shape, tuple) else 1
        self.timesteps = input_shape[0] if isinstance(input_shape, tuple) else input_shape
        self.output = output
        self.lstm = nn.LSTM(self.features, self.output, batch_first=True)
        self.return_sequence = return_sequence
        self.sequential = sequential
        self.hidden = None
        self.cell = None

    def forward(self, x):
        if not self.sequential or self.hidden is None:
            self.hidden = t.zeros(1, x.size(0), self.output).to(next(self.parameters()).device)

        self.cell = t.zeros(1, x.size(0), self.output).to(next(self.parameters()).device)
        out, self.hidden = self.lstm(x, (self.hidden, self.cell))

        if self.return_sequence:
            return out
        else:
            # only return last dimension
            return out[:, -1]


class ResidualLayer(nn.Module):

    def __init__(self, input_size, output_size, activation):
        super().__init__()
        self.l1 = nn.Linear(input_size, output_size)
        self.l2 = nn.Linear(output_size, output_size)
        self.activation = activation

    def forward(self, x):
        x1 = self.activation(self.l1(x))
        x2 = self.activation(self.l2(x1))
        return x1 + x2


class Time2Vec(nn.Module):
    """
    source: https://towardsdatascience.com/time2vec-for-time-series-features-encoding-a03a4f3f937e
    and:    https://arxiv.org/pdf/1907.05321.pdf
    """

    def __init__(self, in_dim: Union[Tuple[int, int], int], out_dim: int):
        super().__init__()
        if not isinstance(in_dim, tuple):
            in_dim = (in_dim, 1)

        self.output_dim = out_dim
        self.w = nn.Parameter(t.Tensor(1, 1))
        self.b = nn.Parameter(t.Tensor(in_dim[0], 1))
        self.W = nn.Parameter(t.Tensor(out_dim * in_dim[1], out_dim * in_dim[1]))
        self.B = nn.Parameter(t.Tensor(in_dim[0], out_dim * in_dim[1]))

    def reset_parameters(self):
        init.uniform_(self.W, 1e-6, 1)
        init.uniform_(self.B, 1e-6, 1)
        init.uniform_(self.w, 1e-6, 1)
        init.uniform_(self.b, 1e-6, 1)

    def forward(self, x):
        if x.ndim < 3:
            x = x.view(*x.shape, 1)

        original = self.w * x + self.b
        x = t.repeat_interleave(x, self.output_dim, dim=-1)
        sin_trans = t.sin((x @ self.W) + self.B)
        return t.cat([sin_trans, original], -1)




from unittest import TestCase

from sklearn.model_selection import KFold
from torch import nn
from torch.optim import SGD

from pandas_ml_common_test.config import TEST_DF
from pandas_ml_utils import FeaturesAndLabels, FittingParameter
from pandas_ml_utils_torch import PytorchNN, PytorchModel
from pandas_ml_utils_torch.merging_cross_folds import take_the_best


class TestCrossFolding(TestCase):

    def test_take_best_model(self):
        df = TEST_DF[["Close"]][-21:].copy()
        with self.assertLogs(level='INFO') as log_ctx:
            with df.model() as m:

                class Net(PytorchNN):

                    def __init__(self):
                        super().__init__()
                        self.net = nn.Sequential(
                            nn.Linear(1, 2),
                            nn.Tanh(),
                            nn.Linear(2, 1),
                            nn.Tanh(),
                        )

                    def forward_training(self, x):
                        return self.net(x)

                fit = m.fit(
                    PytorchModel(
                        Net,
                        FeaturesAndLabels(["Close"], [lambda df: df["Close"].shift(-1)]),
                        nn.MSELoss,
                        lambda params: SGD(params, lr=0.01, momentum=0.0),
                        merge_cross_folds=take_the_best
                    ),
                    FittingParameter(cross_validation=KFold(3), fold_epochs=100, epochs=2)
                )


            print('\n'.join(log_ctx.output))
            self.assertIn('INFO:pandas_ml_utils_torch.merging_cross_folds:best fold: 2!', log_ctx.output)

        self.assertGreater(len(fit.test_summary.df), 0)



import logging

_log = logging.getLogger(__name__)


def reduce(values, reducer, default=None):
    if reducer == 'sum':
        return values.sum()
    elif reducer == 'mean':
        return values.mean()
    else:
        if reducer is not None:
            _log.warning(f"Unknown reduction {reducer}, don't reduce")

        return values if default is None else default



from pandas_ml_utils import FittingParameter
from pandas_ml_utils_test.ml.model.test_model_accuracy import TestModelAccuracy


class TestPytorchModelAccuracy(TestModelAccuracy):

    def provide_linear_regression_model(self):
        from pandas_ml_utils_torch import PytorchModel, PytorchNN
        from pandas_ml_utils import FeaturesAndLabels
        from torch.optim import Adam
        from torch import nn
        import torch as t

        class Net(PytorchNN):

            def __init__(self):
                super(Net, self).__init__()
                self.net = nn.Linear(1, 1)

            def forward_training(self, *input) -> t.Tensor:
                return self.net(input[0])

        return [
            (
                PytorchModel(Net, FeaturesAndLabels(["x"], ["y"]), nn.MSELoss, Adam),
                FittingParameter(epochs=5000, context="epoch fit"),
            ),
            (
                PytorchModel(Net, FeaturesAndLabels(["x"], ["y"]), nn.MSELoss, Adam),
                FittingParameter(epochs=5000, batch_size=64, context="epoch fit batched"),
            ),
            (
                PytorchModel(Net, FeaturesAndLabels(["x"], ["y"]), nn.MSELoss, Adam),
                FittingParameter(epochs=1, fold_epochs=5000, context="fold epoch fit"),
            ),
        ]

    def provide_non_linear_regression_model(self):
        from pandas_ml_utils_torch import PytorchModel, PytorchNN
        from pandas_ml_utils import FeaturesAndLabels
        from torch.optim import Adagrad
        from torch import nn
        import torch as t

        # t.manual_seed(0)

        class Net(PytorchNN):

            def __init__(self):
                super().__init__()
                self.net = nn.Sequential(
                    nn.Linear(1, 200),
                    nn.ReLU(),
                    nn.Linear(200, 200),
                    nn.ReLU(),
                    nn.Linear(200, 200),
                    nn.ReLU(),
                    nn.Linear(200, 1),
                    nn.ReLU()
                )

            def forward_training(self, *input) -> t.Tensor:
                return self.net(input[0])

        t.manual_seed(0)
        model = PytorchModel(Net, FeaturesAndLabels(["x"], ["y"]), nn.MSELoss, Adagrad)

        return [
            (
                model,
                FittingParameter(epochs=600, batch_size=64, context="epoch fit batched"),
            ),
            (
                model,
                FittingParameter(epochs=600, context="epoch fit"),
            ),
            (
                model,
                FittingParameter(epochs=1, fold_epochs=600, context="fold epoch fit"),
            )
        ]






import importlib
from unittest import TestCase

import numpy as np
import pandas as pd
import torch as T
from torch import nn
from torch.distributions import MixtureSameFamily, Categorical, Normal
from torch.optim import Adam

from pandas_ml_common import naive_splitter
from pandas_ml_utils import PostProcessedFeaturesAndLabels, FittingParameter
from pandas_ml_utils.ml.callback import TestConfidenceInterval
from pandas_ml_utils_torch import PytorchModel, PytorchNNFactory, LambdaSplitter
from pandas_ml_utils_torch.loss import DistributionNLL
from pandas_ml_utils_torch.utils import wrap_applyable


class TestProbabilisticModel(TestCase):

    def test_probabilistic_model_with_callback(self):
        try:
            pandas_ml_quant_data_provider = importlib.import_module("pandas_ml_quant")
            from pandas_ml_quant import PricePredictionSummary
            from pandas_ml_quant.model.summary.price_prediction_summary import PriceSampledSummary
        except:
            print("pandas_ml_quant not found, skipping!")
            return

        df = pd.DataFrame({"Returns": np.random.normal(-0.02, 0.03, 500) + np.random.normal(0.03, 0.02, 500)})

        fl = PostProcessedFeaturesAndLabels(
            features=["Returns"],
            feature_post_processor=lambda df: df.ta.rnn(20),
            labels=[
                lambda df: df["Returns"].shift(-1).rename("Future_Returns")
            ],
            targets=lambda df: (1 + df["Returns"]).cumprod().rename("Close")
        )

        model_factory = PytorchNNFactory.create(
            nn.Sequential(
                nn.Linear(20, 10),
                nn.Tanh(),
                nn.Linear(10, 6),
                LambdaSplitter(
                    lambda x: T.softmax(x[..., :2], dim=1),
                    lambda x: T.exp(x[..., 2:4]),
                    # enforce one mean positive and the other negativ
                    lambda x: T.cat([T.exp(x[..., 4:5]), -T.exp(x[..., 5:6])], dim=1),
                )
            ),
            predictor=lambda n, i: T.cat(n(i), dim=1),
            trainer=lambda n, i: n(i)
        )

        def dist(probs, scales, locs):
            return MixtureSameFamily(
                Categorical(probs=probs),
                Normal(loc=locs, scale=scales)
            )

        def loss(y_pred):
            probs, scales, locs = y_pred
            return dist(probs, scales, locs)

        def cdf_cb(arg):
            probs, scales, locs = arg[..., :2], arg[..., 2:4], arg[..., 4:6]
            return dist(probs, scales, locs)

        summary_provider = PriceSampledSummary.with_reconstructor(
            sampler=wrap_applyable(lambda params, samples: cdf_cb(params).sample([int(samples.item())]), nr_args=2),
            samples=100,
            confidence=0.8
        )

        model = PytorchModel(
            module_provider=model_factory,
            features_and_labels=fl,
            criterion_provider=lambda: DistributionNLL(loss, penalize_toal_variance_lambda=1.1),
            optimizer_provider=Adam,
            summary_provider=summary_provider
        )


        fit = df.model.fit(
            model,
            FittingParameter(epochs=10, batch_size=6, splitter=naive_splitter(0.25)),
            #verbose=1,
            callbacks=[
                TestConfidenceInterval(
                    TestConfidenceInterval.CdfConfidenceInterval(
                        wrap_applyable(lambda params, val: cdf_cb(params).cdf(val), nr_args=2),
                        interval=0.8
                    ),
                    wrap_applyable(lambda params: cdf_cb(params).variance),
                    early_stopping=True
                )
            ]
        )

        print(fit.test_summary.calc_scores())



import logging
import sys
from typing import Dict

from pandas_ml_utils_torch.pytorch_base import PytochBaseModel

_log = logging.getLogger(__name__)


def take_the_best(folds: Dict[int, PytochBaseModel]):
    best_loss = sys.float_info.max
    best_module = None
    best_fold = -1

    for fold, model in folds.items():
        _log.info(f'fold: {fold}, loss: {model.last_loss}, best loss: {model.best_loss}')
        if model.last_loss < best_loss:
            best_module = model
            best_fold = fold

    _log.info(f'best fold: {best_fold}!')
    return best_module


def average_folds(folds: Dict[int, PytochBaseModel]):
    mean_model = None

    # add all parameters
    for i, (fold, model) in enumerate(folds.items()):
        if i == 0:
            mean_model = model.shadow_copy()
        else:
            for target_param, local_param in zip(mean_model.net.parameters(), model.net.parameters()):
                target_param.data.copy_(local_param.data + target_param.data)

    # average all parameters
    for target_param in mean_model.net.parameters():
        target_param.data.copy_(target_param.data / len(folds))

    return mean_model





from unittest import TestCase

import torch as t
import torch.nn as nn
from torch.optim import SGD, Adam

from pandas_ml_common import naive_splitter
from pandas_ml_utils import pd, FeaturesAndLabels, FittingParameter
from pandas_ml_utils.ml.model.base_model import AutoEncoderModel
from pandas_ml_utils_test.ml.model.test_abstract_model import TestAbstractModel
from pandas_ml_utils_torch import PytorchModel, PytorchAutoEncoderModel, PytorchNN


class RegressionModule(nn.Module):
    def __init__(self):
        super().__init__()
        self.regressor = nn.Sequential(
            nn.Linear(1, 1)
        )

    def forward(self, x, **kwargs):
        x = self.regressor(x)
        return x


class AutoEncoderModule(PytorchNN):

    def __init__(self):
        super().__init__()
        self.encoder = nn.Sequential(
            nn.Linear(2, 2),
            nn.Tanh(),
            nn.Linear(2, 1),
            nn.Tanh(),
        )

        self.decoder = nn.Sequential(
            nn.Linear(1, 2),
            nn.Tanh(),
            nn.Linear(2, 2),
            nn.Tanh(),
        )

    def forward_training(self, x, **kwargs):
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def encode(self, x):
        return self.encoder(x)

    def decode(self, x):
        return self.decoder(x)


class ClassificationModule(nn.Module):

    def __init__(self):
        super().__init__()
        self.classifier = nn.Sequential(
            nn.Linear(2, 5),
            nn.ReLU(),
            nn.Linear(5, 1),
            nn.Sigmoid()
        )

    def forward(self, x, **kwargs):
        x = self.classifier(x)
        return x


class TestPytorchModel(TestAbstractModel, TestCase):

    def provide_batch_size_and_epoch(self):
        return None, 500

    def provide_classification_model(self, features_and_labels):
        t.manual_seed(42)

        model = PytorchModel(
            ClassificationModule,
            features_and_labels,
            nn.MSELoss,
            lambda params: SGD(params, lr=0.03)
        )

        return model

    def provide_regression_model(self, features_and_labels):
        t.manual_seed(12)

        model = PytorchModel(
            RegressionModule,
            features_and_labels,
            nn.MSELoss,
            lambda params: SGD(params, lr=0.01, momentum=0.0)
        )

        return model

    def provide_auto_encoder_model(self, features_and_labels):
        t.manual_seed(12)

        model = PytorchAutoEncoderModel(
            AutoEncoderModule,
            features_and_labels,
            nn.MSELoss,
            lambda params: SGD(params, lr=0.1, momentum=0.9)
        )

        return model

    def test_mult_epoch_cross_validation(self):
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0, ],
            "b": [0, 1, 0, 1, 1, 0, 1, 0, ],
        })

        with df.model() as m:
            class NN(PytorchNN):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.nn = nn.Sequential(
                        nn.Linear(1, 2),
                        nn.ReLU(),
                        nn.Linear(2, 1),
                    )

                def forward_training(self, x):
                    return self.nn(x)

            fit = m.fit(
                PytorchModel(NN, FeaturesAndLabels(["a"], ["b"]), nn.MSELoss, Adam),
                FittingParameter(
                    splitter=naive_splitter(0.5),
                    epochs=2,
                    fold_epochs=10,
                    batch_size=2
                )
            )

        print(fit)


from .simple_losses import CrossEntropyLoss, MultiObjectiveLoss
from .classificaion import DifferentiableArgmax, ParabolicPenaltyLoss, TailedCategoricalCrossentropyLoss
from .probabilistic import HeteroscedasticityLoss, DistributionNLL
from .soft_dtw import SoftDTW


__version__ = '0.2.7'

from .pytorch_model import PytorchModel, PytorchAutoEncoderModel
from .pytorch_base import PytorchNN, PytorchNNFactory
from .layers import Reshape, KerasLikeLSTM, ResidualLayer, Flatten, Squeeze, LambdaSplitter
from .loss import MultiObjectiveLoss, CrossEntropyLoss, HeteroscedasticityLoss



from unittest import TestCase

import torch as t
import torch.nn as nn
from torch.optim import Adam

from pandas_ml_common.utils.column_lagging_utils import lag_columns
from pandas_ml_common.utils.numpy_utils import one_hot
from pandas_ml_common import np
from pandas_ml_common_test.config import TEST_DF
from pandas_ml_utils import PostProcessedFeaturesAndLabels, FittingParameter
from pandas_ml_utils_torch import PytorchAutoEncoderModel, PytorchNN
from pandas_ml_utils_torch.loss import SoftDTW, TailedCategoricalCrossentropyLoss, ParabolicPenaltyLoss, DifferentiableArgmax


class TestCustomLoss(TestCase):

    def test__differentiable_argmax(self):
        """given"""
        args = 10
        argmax = DifferentiableArgmax(args)

        """when"""

        res = np.array([argmax(t.tensor(one_hot(i, args))).numpy() for i in range(args)])

        """then"""
        print(res)
        np.testing.assert_array_almost_equal(res, np.arange(0, args))

    def test__parabolic_crossentropy(self):
        """when"""
        categories = 11
        loss_function = ParabolicPenaltyLoss(categories, 1)

        """then"""
        for truth in range(categories):
            losses = []
            for prediction in range(categories):
                loss = loss_function(t.tensor(one_hot(prediction, categories)),
                                     t.tensor(one_hot(truth, categories)))
                losses.append(loss)

            # all predictions left of truth need to increase
            for i in range(1, truth):
                self.assertGreater(losses[i - 1], losses[i])

            # right of truth need to decrease
            for i in range(truth, categories - 1):
                self.assertLess(losses[i], losses[i + 1])

            if truth > 0 and truth < categories - 1:
                if truth > categories / 2:
                    # right tail:
                    self.assertGreater(losses[truth - 1], losses[truth + 1])
                else:
                    # left tail
                    self.assertGreater(losses[truth + 1], losses[truth - 1])

    def test__tailed_categorical_crossentropy(self):
        """when"""
        categories = 11
        loss = TailedCategoricalCrossentropyLoss(categories, 1)

        """then"""
        truth = one_hot(3, 11)
        l = loss(t.tensor([one_hot(6, 11)]), t.tensor([truth]))
        l2 = loss(t.tensor([one_hot(6, 11), one_hot(9, 11)]), t.tensor([truth, truth]))

        """then"""
        np.testing.assert_almost_equal(l, 11.817837, decimal=5)
        np.testing.assert_array_almost_equal(l2, [11.817837, 42.46247], decimal=5)
        self.assertGreater(l.mean().numpy(), 0)

    def test__tailed_categorical_crossentropy_3d(self):
        loss = TailedCategoricalCrossentropyLoss(11, 1)
        self.assertEqual(loss(t.ones((32, 11), requires_grad=True), t.ones((32, 1, 11), requires_grad=True)).shape,
                         (32,))
        self.assertEqual(loss(t.ones((32, 1, 11), requires_grad=True), t.ones((32, 1, 11), requires_grad=True)).shape,
                         (32,))
        self.assertEqual(loss(t.ones((32, 11), requires_grad=True), t.ones((32, 11), requires_grad=True)).shape,
                         (32,))

    def test_soft_dtw_loss(self):
        df = TEST_DF[["Close"]][-21:].copy()

        class LstmAutoEncoder(PytorchNN):
            def __init__(self):
                super().__init__()
                self.input_size = 1
                self.seq_size=10
                self.hidden_size = 2
                self.num_layers = 1
                self.num_directions = 1

                self._encoder =\
                    nn.RNN(input_size=self.input_size, hidden_size=self.hidden_size, num_layers=self.num_layers,
                           batch_first=True)

                self._decoder =\
                    nn.RNN(input_size=self.hidden_size, hidden_size=self.input_size, num_layers=self.num_layers,
                           batch_first=True)

            def forward_training(self, x):
                # make sure to treat single elements as batches
                x = x.view(-1, self.seq_size, self.input_size)
                batch_size = len(x)

                hidden_encoder = nn.Parameter(t.zeros(self.num_layers * self.num_directions, batch_size, self.hidden_size))
                hidden_decoder = nn.Parameter(t.zeros(self.num_layers * self.num_directions, batch_size, self.input_size))

                x, _ = self._encoder(x, hidden_encoder)
                x = t.repeat_interleave(x[:,-2:-1], x.shape[1], dim=1)
                x, hidden = self._decoder(x, hidden_decoder)
                return x.squeeze()

            def encode(self, x):
                x = x.reshape(-1, self.seq_size, self.input_size)
                batch_size = len(x)

                with t.no_grad():
                    hidden = nn.Parameter(
                        t.zeros(self.num_layers * self.num_directions, batch_size, self.hidden_size))

                    # return last element of sequence
                    return self._encoder(x, hidden)[0][:,-1]

            def decode(self, x):
                x = x.reshape(-1, self.seq_size, self.hidden_size)
                batch_size = len(x)

                with t.no_grad():
                    hidden = nn.Parameter(
                        t.zeros(self.num_layers * self.num_directions, batch_size, self.input_size))
                    return self._decoder(x.float(), hidden)[0]

        model = PytorchAutoEncoderModel(
            LstmAutoEncoder,
            PostProcessedFeaturesAndLabels(df.columns.to_list(), [lambda df: lag_columns(df, 10).dropna()],
                                           df.columns.to_list(), [lambda df: lag_columns(df, 10).dropna()],
                                           ["condensed-a", "condensed-b"]),
            SoftDTW,
            Adam
        )

        with df.model() as m:
            fit = m.fit(model, FittingParameter(epochs=100))
            print(fit.test_summary.df)

            encoded = df.model.predict(fit.model.as_encoder())
            print(encoded)



import unittest
from unittest import TestCase

import torch.nn as nn
from torch.optim import Adam

from pandas_ml_common.decorator import MultiFrameDecorator
from pandas_ml_utils import pd
from pandas_ml_utils.ml.data.extraction.features_and_labels_extractor import FeaturesWithLabels
from pandas_ml_utils_torch import PytorchModel, PytorchNN
from pandas_ml_utils import FeaturesAndLabels
from pandas_ml_utils.constants import FEATURE_COLUMN_NAME
import torch as t
import numpy as np


class TestMultiFeatureSet(TestCase):

    def test_pytorch_mfs(self):
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [[0, 0], [0, 0], [1, 1], [1,1], [0,0], [0,0], [1,1], [1,1],],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,]
        })

        def module_provider():
            class ClassificationModule(PytorchNN):

                def __init__(self):
                    super().__init__()
                    self.net0 = nn.Sequential(
                        nn.Linear(1, 5),
                        nn.ReLU(),
                        nn.Linear(5, 1),
                        nn.Sigmoid()
                    )
                    self.net1 = nn.Sequential(
                        nn.Linear(2, 5),
                        nn.ReLU(),
                        nn.Linear(5, 1),
                        nn.Sigmoid()
                    )

                def forward_training(self, x) -> t.Tensor:
                    x0, x1 = x
                    return self.net0(x0) + self.net1(x1)

            return ClassificationModule()

        model = PytorchModel(
            module_provider,
            FeaturesAndLabels(
                features=(["a"], ["b"]),
                labels=["c"]
            ),
            nn.MSELoss,
            lambda params: Adam(params, lr=0.03)
        )

        fl: FeaturesWithLabels = df._.extract(model.features_and_labels)
        self.assertIsInstance(fl.features_with_required_samples.features, MultiFrameDecorator)
        print(fl.features_with_required_samples.features)

        fit = df.model.fit(model, fold_epochs=10)
        print(fit.test_summary.df)

        self.assertIn(FEATURE_COLUMN_NAME, fit.test_summary.df)
        np.testing.assert_almost_equal(np.array([0, 0, 1]), fit.test_summary.df["label"].values.squeeze())


from numbers import Number
from typing import Tuple, Union, Iterable

import numpy as np

from pandas_ml_common import pd
import torch as t

from pandas_ml_common.decorator import MultiFrameDecorator
from pandas_ml_common.utils import unpack_nested_arrays


def from_pandas(df: pd.DataFrame, cuda: bool = False, default: t.Tensor = None) -> Union[t.Tensor, Tuple[t.Tensor, ...]]:
    if df is None and default is None:
        return None

    if isinstance(df, MultiFrameDecorator):
        return tuple([from_pandas(f, cuda, default) for f in df.frames()])

    val = (t.from_numpy(df._.get_values(split_multi_index_rows=False, squeeze=True)) if df is not None else default).float()
    return val.cuda() if cuda else val.cpu()


def to_device(var, cuda):
    return var.cuda() if cuda else var.cpu()


def copy_weights(source_network, target_network):
    target_network.load_state_dict(source_network.state_dict())
    return target_network.train(source_network.training)


def wrap_applyable(func, nr_args=1, return_numpy=True):
    def wrapped(cells):
        x = []
        for i in range(nr_args):
            cell = (cells.iloc[i] if hasattr(cells, 'iloc') else cells[i]) if nr_args > 1 else cells
            if isinstance(cell, (np.ndarray, pd.Series, pd.DataFrame, Number, list, tuple, set)):
                if hasattr(cell, 'item') and sum(cell.shape) <= 1:
                    x.append(t.Tensor(unpack_nested_arrays(cell.item()) if cell.dtype == 'object' else [cell.item()]))
                elif hasattr(cell, 'values'):
                    x.append(t.Tensor(unpack_nested_arrays(cell.values) if cell.dtype == 'object' else cell.values))
                else:
                    x.append(t.Tensor(cell if isinstance(cell, Iterable) else [cell]))
            else:
                x.append(cell)

        x = func(*x)
        return (x.numpy() if sum(x.shape) > 1 else x.item()) if return_numpy else x

    return wrapped



from torch.distributions import Uniform, SigmoidTransform, AffineTransform, TransformedDistribution
from torch.distributions import constraints
import torch as t


class Logistic(TransformedDistribution):
    r"""
    Creates a logistic distribution parameterized by :attr:`loc` and :attr:`scale`
    that define the base `Uniform` distribution transformed with the
    `SigmoidTransform` and `AffineTransform` such that:
        X ~ Logistic(loc, scale)
        cdf(x; loc, scale) = 1 / (1 + exp(-(x - loc) / scale))
    Args:
        loc (float or Tensor): mean of the base distribution
        scale (float or Tensor): standard deviation of the base distribution
    Example::
        >>> # logistic-normal distributed with mean=(0, 0, 0) and stddev=(1, 1, 1)
        >>> # of the base Normal distribution
        >>> m = distributions.Logistic(torch.tensor([0.0] * 3), torch.tensor([1.0] * 3))
        >>> m.sample()
        tensor([ 1.2082, -3.3503,  0.0038])
    """
    arg_constraints = {'loc': constraints.real, 'scale': constraints.positive}
    support = constraints.simplex
    has_rsample = True

    def __init__(self, loc, scale, validate_args=None):
        base_dist = Uniform(t.zeros(loc.shape), t.ones(loc.shape), validate_args=validate_args)
        if not base_dist.batch_shape:
            base_dist = base_dist.expand([1])
        super(Logistic, self).__init__(
            base_dist,
            [SigmoidTransform().inv, AffineTransform(loc, scale)]
        )

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(Logistic, _instance)
        return super(Logistic, self).expand(batch_shape, _instance=new)

    @property
    def loc(self):
        return self.transforms[-1].loc

    @property
    def scale(self):
        return self.transforms[-1].scale



from typing import Callable, Any

import torch as t
import logging

from torch.distributions import Distribution

from ._loss_utils import reduce

PI = t.acos(t.zeros(1)) * 2
_log = logging.getLogger(__name__)


class DistributionNLL(t.nn.Module):

    def __init__(self, distribution_provider: Callable[[Any], Distribution], reduction: str = 'sum', penalize_toal_variance_lambda: float = None):
        super().__init__()
        self.distribution_provider = distribution_provider
        self.penalize_toal_variance = penalize_toal_variance_lambda
        self.reduction = reduction

    def forward(self, y_pred, y_true):
        dist = self.distribution_provider(y_pred)
        penalty = 0 if self.penalize_toal_variance is None else dist.variance * self.penalize_toal_variance
        return reduce(-dist.log_prob(y_true) + penalty, self.reduction)


class HeteroscedasticityLoss(t.nn.Module):

    def __init__(self, reduction=None, multi_nominal_reduction='sum'):
        super().__init__()
        self.reduction = reduction
        self.multi_nominal_reduction = multi_nominal_reduction

    def forward(self, y_pred, y_true):
        mu = t.narrow(y_pred, 1, 0, 1)
        sigma = t.exp(t.narrow(y_pred, 1, 1, 1))

        a = 1 / (t.sqrt(2 * PI) * sigma)
        b1 = t.square(mu - y_true)
        b2 = 2 * t.square(sigma)
        b = b1 / b2

        loss = (-t.log(a) + b).squeeze()

        if loss.ndim > 1:
            if self.multi_nominal_reduction == 'sum':
                loss = loss.sum(dim=tuple(range(1, loss.ndim)))
            elif self.multi_nominal_reduction == 'mean':
                loss = loss.mean(dim=tuple(range(1, loss.ndim)))
            elif isinstance(self.multi_nominal_reduction, (tuple, list)):
                raise NotImplementedError("Not yet implemented, sorry :-(")
            else:
                if self.reduction is not None:
                    _log.warning(f"Unknown multi nominal reduction {self.multi_nominal_reduction}, need to be one of ('sum', 'mean', a list of weigts as floats)")

        return reduce(loss, self.reduction)




from typing import Dict
from unittest import TestCase

import torch as T
import torch.nn as nn
from torch.optim import Adam

from pandas_ml_utils import FeaturesAndLabels, FittingParameter
from pandas_ml_utils_torch import PytorchModel, PytorchNN
from pandas_ml_utils import pd, np
from pandas_ml_common.sampling import naive_splitter
from pandas_ml_utils_torch.loss import MultiObjectiveLoss


class TestLoss(TestCase):

    def test_multi_objective_loss(self):
        df = pd.DataFrame(np.array([
            # train
            [0, 0, 0],
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 1],
            # test
            [0, 0, 0],
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 1],
        ]), columns=["f1", "f2", "l"])

        class XorModule(PytorchNN):

            def __init__(self):
                super().__init__()
                self.x1 = nn.Linear(2, 1)
                self.s1 = nn.Sigmoid()
                self.x2 = nn.Linear(2, 1)
                self.s2 = nn.Sigmoid()
                self.s = nn.Softmax()

            def forward_training(self, x):
                return self.s1(self.x1(x)), self.s2(self.x2(x))

            def forward_predict(self, x):
                return self.s1(self.x1(x))


        fit = df.model.fit(
            PytorchModel(
                XorModule,
                FeaturesAndLabels(["f1", "f2"], ["l"]),
                lambda: MultiObjectiveLoss((1, nn.MSELoss(reduction='none')),
                                           (1, nn.L1Loss(reduction='none')),
                                           on_epoch=lambda criterion, epoch: criterion.update_weights((0, 1.1))),
                Adam
            ),
            FittingParameter(splitter=naive_splitter(0.5))
        )

        print(fit.test_summary.df)

    def test_regularized_loss(self):
        df = pd.DataFrame({
            "f": np.sin(np.linspace(0, 12, 40)),
            "l": np.sin(np.linspace(5, 17, 40))
        })

        class TestModel(PytorchNN):

            def __init__(self):
                super().__init__()
                self.net = nn.Sequential(
                    nn.Linear(1, 3),
                    nn.ReLU(),
                    nn.Linear(3, 2),
                    nn.ReLU(),
                    nn.Linear(2, 1),
                    nn.Sigmoid()
                )

            def forward_training(self, x):
                return self.net(x)

            def L2(self) -> Dict[str, float]:
                return {
                    '**/2/**/weight': 99999999999.99
                }

        fit = df.model.fit(
            PytorchModel(
                TestModel,
                FeaturesAndLabels(["f"], ["l"]),
                nn.MSELoss,
                Adam
            ),
            FittingParameter(epochs=1000, splitter=naive_splitter(0.5))
        )

        print(fit.model._current_model.net.net[2].weight.detach().numpy())
        print(fit.model._current_model.net.net[2].weight.norm().detach().item())
        self.assertLess(fit.model._current_model.net.net[2].weight.norm().detach().item(), 0.1)

    def test_probabilistic(self):
        def create_sine_data(n=300):
            np.random.seed(32)
            n = 300
            x = np.linspace(0, 1 * 2 * np.pi, n)
            y1 = 3 * np.sin(x)
            y1 = np.concatenate((np.zeros(60), y1 + np.random.normal(0, 0.15 * np.abs(y1), n), np.zeros(60)))
            x = np.concatenate((np.linspace(-3, 0, 60), np.linspace(0, 3 * 2 * np.pi, n),
                                np.linspace(3 * 2 * np.pi, 3 * 2 * np.pi + 3, 60)))
            y2 = 0.1 * x + 1
            y = y1 + y2
            return x, y

        df = pd.DataFrame(np.array(create_sine_data(300)).T, columns=["x", "y"])
        with df.model() as m:
            from pandas_ml_utils import FeaturesAndLabels
            from pandas_ml_utils_torch import PytorchNN, PytorchModel
            from pandas_ml_utils_torch.loss import HeteroscedasticityLoss
            from pandas_ml_common.sampling.splitter import duplicate_data
            from torch.optim import Adam
            from torch import nn

            class Net(PytorchNN):

                def __init__(self):
                    super().__init__()
                    self.l = nn.Sequential(
                        nn.Linear(1, 20),
                        nn.ReLU(),
                        nn.Linear(20, 50),
                        nn.ReLU(),
                        nn.Linear(50, 20),
                        nn.ReLU(),
                        nn.Linear(20, 2),
                    )

                def forward_training(self, x):
                    return self.l(x)

            fit = m.fit(
                PytorchModel(
                    Net,
                    FeaturesAndLabels(["x"], ["y"]),
                    HeteroscedasticityLoss,
                    Adam,
                    restore_best_weights=True
                ),
                FittingParameter(batch_size=128, epochs=10, splitter=duplicate_data())
            )

    def test_heteroscedasticity_loss_3d(self):
        from pandas_ml_utils_torch.loss import HeteroscedasticityLoss

        y_true = T.ones((10, 1, 3))
        y_pred = T.randn((10, 2, 3))
        critereon = HeteroscedasticityLoss(multi_nominal_reduction=None)

        t3d = critereon(y_pred, y_true)
        t2d = critereon(y_pred[:, :, 0], y_true[:, :, 0])

        print(t3d, '\n', t2d)
        np.testing.assert_array_equal(t2d.numpy(), t3d[:, 0].numpy())



from unittest import TestCase

import numpy as np
import numpy.testing
import pandas as pd

from pandas_ml_utils_torch.utils import wrap_applyable


class TestUtils(TestCase):

    def test_wrap_applyable(self):
        df = pd.DataFrame({
            "a": [1, 2, 3],
            "b": [[1, 2, 3],[1, 2, 3],[1, 2, 3]],
            "c": [[[1, 2, 3]],[[1, 2, 3]],[[1, 2, 3]]],
        })

        np.testing.assert_array_almost_equal(
            df[["a"]].apply(wrap_applyable(lambda x: x*1), axis=1, raw=False),
            [1, 2, 3]
        )

        np.testing.assert_array_almost_equal(
            df[["a"]].apply(wrap_applyable(lambda x: x*1), axis=1, raw=True),
            [1, 2, 3]
        )

        np.testing.assert_array_almost_equal(
            df[["a", "a"]].apply(wrap_applyable(lambda x: x.sum()), axis=1, raw=False),
            [2, 4, 6]
        )

        np.testing.assert_array_almost_equal(
            df[["a", "a"]].apply(wrap_applyable(lambda x: x.sum()), axis=1, raw=True),
            [2, 4, 6]
        )

        np.testing.assert_array_almost_equal(
            df["a"].apply(wrap_applyable(lambda x: x*1)),
            [1, 2, 3]
        )

        np.testing.assert_array_almost_equal(
            df["b"].apply(wrap_applyable(lambda x: sum(x.shape), return_numpy=False)).values,
            [3, 3, 3]
        )

        np.testing.assert_array_almost_equal(
            df[["c"]].apply(wrap_applyable(lambda x: sum(x.shape), return_numpy=False), axis=1).values,
            [4, 4, 4]
        )

        np.testing.assert_array_almost_equal(
            df[["c"]].apply(wrap_applyable(lambda x: sum(x.shape), return_numpy=False), axis=1, raw=True).values,
            [4, 4, 4]
        )

    def test_wrap_distribution(self):
        from torch.distributions import MixtureSameFamily, Categorical, Normal

        df = pd.DataFrame({
            "b": [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
            "c": [1, 2, 3],
        })

        def variant_a(tens):
            return MixtureSameFamily(
                Categorical(probs=tens[:1]),
                Normal(loc=tens[1:2], scale=tens[2:3])
            )

        def variant_b(tens, b):
            return MixtureSameFamily(
                Categorical(probs=tens[:1]),
                Normal(loc=tens[1:2], scale=tens[2:3])
            ).cdf(b)

        # variant a
        dists_a = df[["c"]]\
            .join(df["b"].apply(wrap_applyable(variant_a, return_numpy=False)))\
            .apply(wrap_applyable(lambda val, dist: dist.cdf(val), nr_args=2), axis=1)

        # variant b
        dists_b = df.apply(wrap_applyable(variant_b, nr_args=2), axis=1)

        numpy.testing.assert_array_almost_equal(dists_a, dists_b)


from unittest import TestCase
from torch.optim import Adam
from torch import nn
import torch

from pandas_ml_utils_torch.merging_cross_folds import average_folds
from pandas_ml_utils_torch.pytorch_base import PytochBaseModel, PytorchNN


class TestMergeFolds(TestCase):

    def test_average_merge(self):
        class Net(PytorchNN):

            def __init__(self, init_weight):
                super().__init__()
                self.l1 = nn.Linear(1, 1)
                torch.nn.init.constant_(self.l1.weight, init_weight)

            def forward(self, x):
                return self.l1(x)


        a = PytochBaseModel(lambda: Net(1), nn.MSELoss, Adam)
        b = PytochBaseModel(lambda: Net(2), nn.MSELoss, Adam)

        res = list(average_folds({1: a, 2: b}).net.parameters())
        self.assertAlmostEqual(res[0].data.numpy().item(), 1.5)
